"""
Apply ``func(*args, value, **kwargs)`` to each emitted value. It's also
possible to omit ``func`` - in this case it's acting as dummy subscriber

Usage:

>>> from broqer import Value, op, Sink
>>> s = Value()

>>> len(s.subscriptions)
0
>>> _d = s.subscribe(Sink(print, 'Sink', sep=':'))
>>> len(s.subscriptions)
1

>>> s.emit(1)
Sink:1
>>> s.emit((1, 2))
Sink:(1, 2)

>>> _d.dispose()
>>> len(s.subscriptions)
0
"""
from functools import partial, wraps
from typing import Any, Callable, Optional, TYPE_CHECKING

from broqer import Subscriber

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from broqer import Publisher


class Sink(Subscriber):  # pylint: disable=too-few-public-methods
    """ Apply ``function(*args, value, **kwargs)`` to each emitted value. It's
    also possible to omit ``function`` - in this case it's acting as dummy
    subscriber

    :param function: function to be called when source publisher emits
    :param \\*args: variable arguments to be used for calling function
    :param unpack: value from emits will be unpacked (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for calling function
    """
    def __init__(self,  # pylint: disable=keyword-arg-before-vararg
                 function: Optional[Callable[..., None]] = None,
                 *args, unpack=False, **kwargs) -> None:
        if function is None:
            self._function = None  # type: Optional[Callable[..., None]]
        elif args or kwargs:
            self._function = \
                partial(function, *args, **kwargs)
        else:
            self._function = function

        self._unpack = unpack

    def emit(self, value: Any, who: 'Publisher'):
        if self._function:
            if self._unpack:
                self._function(*value)
            else:
                self._function(value)


def build_sink(function: Callable[..., None] = None, *,
               unpack: bool = False):
    """ Decorator to wrap a function to return a Sink subscriber.

    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_sink(function):
        return Sink(function, unpack=unpack)

    if function:
        return _build_sink(function)

    return _build_sink


def build_sink_factory(function: Callable[..., None] = None, *,
                       unpack: bool = False):
    """ Decorator to wrap a function to return a Sink subscriber factory.
    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_sink(function: Callable[..., None]):
        @wraps(function)
        def _wrapper(*args, **kwargs) -> Sink:
            if 'unpack' in kwargs:
                raise TypeError('"unpack" has to be defined by decorator')
            return Sink(function, *args, unpack=unpack, **kwargs)
        return _wrapper

    if function:
        return _build_sink(function)

    return _build_sink


def sink_property(function: Callable[..., None] = None, unpack: bool = False):
    """ Decorator to build a property returning a Sink subscriber.
    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def build_sink_property(function):
        @property
        def _build_sink(self):
            return Sink(function, self, unpack=unpack)
        return _build_sink

    if function:
        return build_sink_property(function)

    return build_sink_property



# -*- coding: utf-8 -*-
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

project = 'broqer'
copyright = u'2018, Günther Jena'
author = u'Günther Jena'
from broqer import __version__ as version
github_url = 'https://github.com/semiversus/python-broqer'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.viewcode',
]

master_doc = 'index'
source_suffix = '.rst'

html_theme = 'sphinx_rtd_theme'
htmlhelp_basename = 'broqer'

html_context = {
  'display_github': True, # Add 'Edit on Github' link instead of 'View page source'
  'github_user': 'semiversus',
  'github_repo': project,
  'github_version': 'master',
  'conf_py_path': '/docs/',
  'source_suffix': source_suffix,
}

latex_documents = [
    (master_doc, 'python-broqer.tex', 'python-broqer Documentation',
     u'Günther Jena', 'manual'),
]

from sphinx.ext import autodoc

class ClassDocDocumenter(autodoc.ClassDocumenter):
    objtype = 'classdoc'

    #do not indent the content
    content_indent = ""

    #do not add a header to the docstring
    def add_directive_header(self, sig):
        pass

def setup(app):
    app.add_autodocumenter(ClassDocDocumenter)



from unittest import mock

from broqer import Publisher, NONE, Sink
from broqer.op import Concat, Map


def test_operator_concat():
    DUT = Concat(Map(lambda v: v/2), Map(lambda v: v+1))
    mock_cb = mock.Mock()

    p = Publisher()

    o = p | DUT
    assert o.get() == NONE

    p.notify(0)
    assert o.get() == 1.0

    o.subscribe(Sink(mock_cb))

    for v in range(5):
        p.notify(v)

    mock_cb.assert_has_calls(
        [mock.call(1.0), mock.call(1.5), mock.call(2.0), mock.call(2.5),
         mock.call(3.0)])

    assert o.get() == 3.0



""" Broqer is a carefully crafted library to operate with continuous streams
of data in a reactive style with publish/subscribe and broker functionality.
"""

from .error_handler import default_error_handler
from .disposable import Disposable
from .types import NONE
from .publisher import Publisher, SubscriptionDisposable, SubscriptionError
from .subscriber import Subscriber
from .subscribers import (OnEmitFuture, Sink, Trace, build_sink,
                          build_sink_factory, sink_property, SinkAsync,
                          build_sink_async, build_sink_async_factory,
                          sink_async_property)
from .value import Value

from .operator_overloading import apply_operator_overloading

apply_operator_overloading()


__author__ = 'Günther Jena'
__email__ = 'guenther@jena.at'

try:
    from ._version import version as __version__
except ImportError:
    __version__ = 'not available'

__all__ = [
    'default_error_handler', 'Disposable', 'NONE', 'Publisher',
    'SubscriptionDisposable', 'SubscriptionError', 'Subscriber',
    'OnEmitFuture', 'Sink', 'Trace', 'build_sink', 'build_sink_factory',
    'sink_property', 'Value', 'op', 'SinkAsync', 'build_sink_async',
    'build_sink_async_factory', 'sink_async_property'
]



"""
Rate limit emits by the given time.
Usage:
>>> import asyncio
>>> from broqer import Value, op, Sink
>>> v = Value()
>>> throttle_publisher = v | op.Throttle(0.1)
>>> _d = throttle_publisher.subscribe(Sink(print))
>>> v.emit(1)
1
>>> v.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
>>> v.emit(3)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.2))
3
>>> # It's also possible to reset the throttling duration:
>>> v.emit(4)
4
>>> v.emit(5)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
>>> throttle_publisher.reset()
"""
import asyncio
import sys
from typing import Any  # noqa: F401

from broqer import Publisher, default_error_handler, NONE

from broqer.operator import Operator, OperatorFactory


class AppliedThrottle(Operator):
    """ Rate limit emits by the given time.
    :param duration: time for throttling in seconds
    :param error_callback: the error callback to be registered
    :param loop: asyncio event loop to use
    """
    def __init__(self, publisher: Publisher, duration: float,
                 error_callback=default_error_handler, loop=None) -> None:

        Operator.__init__(self, publisher)

        self._duration = duration
        self._loop = loop or asyncio.get_event_loop()
        self._call_later_handler = None  # type: asyncio.Handle
        self._last_state = NONE  # type: Any
        self._error_callback = error_callback

    def get(self):
        return Publisher.get(self)

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if self._call_later_handler is None:
            self._last_state = value
            self._wait_done_cb()
        else:
            self._last_state = value

    def _wait_done_cb(self):
        if self._last_state is not NONE:
            try:
                Publisher.notify(self, self._last_state)
            except Exception:  # pylint: disable=broad-except
                self._error_callback(*sys.exc_info())
            self._last_state = NONE
            self._call_later_handler = self._loop.call_later(
                self._duration, self._wait_done_cb)
        else:
            self._call_later_handler = None

    def reset(self):
        """ Reseting duration for throttling """
        if self._call_later_handler is not None:
            self._call_later_handler.cancel()
            self._call_later_handler = None


class Throttle(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Apply throttling to each emitted value.
    :param duration: time for throttling in seconds
    """
    def __init__(self, duration: float) -> None:
        if duration < 0:
            raise ValueError('Duration has to be bigger than zero')

        self._duration = duration

    def apply(self, publisher: Publisher):
        return AppliedThrottle(publisher, self._duration)



""" The op module contains all operators broqer offers """

# synchronous operators
from broqer.op.combine_latest import CombineLatest, build_combine_latest
from broqer.op.filter_ import Filter, EvalTrue, EvalFalse, build_filter, \
                              build_filter_factory
from broqer.op.map_ import Map, build_map, build_map_factory
from broqer.op.map_async import MapAsync, build_map_async, \
                                build_map_async_factory, AsyncMode
from broqer.op.bitwise import BitwiseCombineLatest, map_bit

# utils
from broqer.op.concat import Concat
from broqer.op.throttle import Throttle

# enable operator overloading
from .py_operators import Str, Bool, Int, Float, Repr, Len, In, All, Any, \
                         BitwiseAnd, BitwiseOr, Not

__all__ = [
    'CombineLatest', 'BitwiseCombineLatest',
    'Filter', 'Map', 'EvalTrue', 'MapAsync', 'build_map_async', 'AsyncMode',
    'EvalFalse', 'build_map', 'build_map_factory', 'build_combine_latest',
    'build_filter', 'build_filter_factory', 'Concat', 'Str', 'Bool', 'Int',
    'Float', 'Repr', 'map_bit', 'build_map_async_factory',
    'Len', 'In', 'All', 'Any', 'BitwiseAnd', 'BitwiseOr', 'Not', 'Throttle'
]



from unittest import mock

import pytest

from broqer import Publisher, SubscriptionError, Value, NONE, Sink


def test_subscribe():
    """ Testing .subscribe() and .unsubscibe() of Publisher. """
    s1 = Value()
    s2 = Value()

    publisher = Publisher()
    assert len(publisher.subscriptions) == 0

    # subscribe first subscriber
    d1 = publisher.subscribe(s1)
    assert any(s1 is s for s in publisher.subscriptions)
    assert not any(s2 is s for s in publisher.subscriptions)
    assert len(publisher.subscriptions) == 1

    # re - subscribe should fail
    with pytest.raises(SubscriptionError):
        publisher.subscribe(s1)

    # subscribe second subscriber
    d2 = publisher.subscribe(s2)
    assert len(publisher.subscriptions) == 2
    assert any(s1 is s for s in publisher.subscriptions)
    assert any(s2 is s for s in publisher.subscriptions)

    # unsubscribe both subscribers
    d2.dispose()
    assert len(publisher.subscriptions) == 1
    publisher.unsubscribe(s1)
    assert len(publisher.subscriptions) == 0

    # re - unsubscribing should fail
    with pytest.raises(SubscriptionError):
        d1.dispose()

    with pytest.raises(SubscriptionError):
        publisher.unsubscribe(s1)

    with pytest.raises(SubscriptionError):
        d2.dispose()


@pytest.mark.asyncio
async def test_await(event_loop):
    """ Test .as_future() method """
    publisher = Publisher()

    # test omit_subscription=False when publisher has no state. This should
    # wait for the first state change
    event_loop.call_soon(publisher.notify, 1)
    assert await publisher.as_future(timeout=1, omit_subscription=False) == 1

    # omit_subscription is defaulted to True, so the current state should not
    # be returned, instead it should wait for the first change
    event_loop.call_soon(publisher.notify, 2)
    assert await publisher.as_future(timeout=1) == 2

    # when publisher has a state and omit_subscription is False it should
    # directly return publisher's state
    event_loop.call_soon(publisher.notify, 3)
    assert await publisher.as_future(timeout=1, omit_subscription=False) == 2

    assert (await publisher) == 2


@pytest.mark.parametrize('init', [0, 'Test', {'a':1}, None, (1,2,3), [1,2,3]])
def test_get(init):
    """ Testing .get() method """
    # testing .get() after initialisation
    p1 = Publisher(init)
    p2 = Publisher()
    p3 = Publisher(3)

    assert p1.get() is init
    assert p2.get() is NONE
    assert p3.get() is 3

    # testing .get() after notfiy
    p1.notify(1)
    p2.notify(init)
    p3.notify(init)

    assert p1.get() is 1
    assert p2.get() is init
    assert p3.get() is init


@pytest.mark.parametrize('number_of_subscribers', range(3))
@pytest.mark.parametrize('init', [NONE, 0, 'Test', {'a':1}, None, (1,2,3), [1,2,3]])
def test_notify(init, number_of_subscribers):
    """ Testing .notify(v) method """
    p = Publisher(init)
    m = mock.Mock()

    # subscribing Sinks to the publisher and test .notify on subscribe
    subscribers = [Sink(m, i) for i in range(number_of_subscribers)]

    m.assert_not_called()

    for s in subscribers:
        p.subscribe(s)

    if init is not NONE:
        m.assert_has_calls([mock.call(i, init) for i in range(number_of_subscribers)])
    else:
        m.assert_not_called()

    m.reset_mock()

    # test .notify() for listening subscribers
    p.notify(1)

    m.assert_has_calls([mock.call(i, 1) for i in range(number_of_subscribers)])

    m.reset_mock()

    # re- .notify() with the same value
    p.notify(1)

    m.assert_has_calls([mock.call(i, 1) for i in range(number_of_subscribers)])


def test_subscription_callback():
    """ testing .register_on_subscription_callback() """
    m = mock.Mock()

    p = Publisher()
    p.register_on_subscription_callback(m)

    # callback should be called on first subscription (with True as argument)
    d1 = p.subscribe(Sink())
    m.assert_called_once_with(True)
    m.reset_mock()

    # it should not be called until the last subscriber is unsubscribing
    d2 = p.subscribe(Sink())
    d1.dispose()

    m.assert_not_called()

    # callback should be called with False as argument on last unsubscription
    d2.dispose()
    m.assert_called_once_with(False)
    m.reset_mock()

    # callback should again be called with True as argument on first subscription
    d3 = p.subscribe(Sink())
    m.assert_called_once_with(True)
    m.reset_mock()

    # after reseting callback it should no be called again
    p.register_on_subscription_callback(None)
    d3.dispose()
    p.subscribe(Sink())
    m.assert_not_called()

    # check if callback is called when subscriber is already available
    p.register_on_subscription_callback(m)
    m.assert_called_once_with(True)




def test_prepend():
    """ Testing the prepend argument in .subscribe() """
    m = mock.Mock()
    p = Publisher()

    s1 = Sink(m, 1)
    s2 = Sink(m, 2)
    s3 = Sink(m, 3)
    s4 = Sink(m, 4)

    p.subscribe(s1)
    p.subscribe(s2, prepend=True)  # will be inserted before s1
    p.subscribe(s3)  # will be appended after s1
    p.subscribe(s4, prepend=True)  # will be inserted even before s2

    p.notify('test')

    # notification order should now be: s4, s2, s1, s3
    m.assert_has_calls([mock.call(4, 'test'), mock.call(2, 'test'), mock.call(1, 'test'), mock.call(3, 'test')])


def test_reset_state():
    """ Test .reset_state() """
    m = mock.Mock()
    p = Publisher()

    p.subscribe(Sink(m, 1))

    m.assert_not_called()

    # test .reset_state() before and after subscribing
    p.reset_state()
    assert p.get() == NONE

    p.subscribe(Sink(m, 2))

    m.assert_not_called()

    m.reset_mock()

    # test .reset_state() after notify
    p.notify('check')
    assert p.get() == 'check'
    m.assert_has_calls([mock.call(1, 'check'), mock.call(2, 'check')])
    m.reset_mock()

    # test no subscribers get notified
    p.reset_state()
    m.assert_not_called()

    assert p.get() == NONE



import asyncio
import statistics

from broqer import op
from broqer.subject import Subject

adc_raw = Subject()

(adc_raw
 | op.Cache(0)
 | op.Map(lambda d: d * 5 + 3)
 | op.Sample(0.3)
 | op.SlidingWindow(5)
 | op.Map(statistics.mean)
 | op.Cache()
 | op.Debounce(0.5)
 | op.Sink(print)
 )


async def main():
    await asyncio.sleep(2)
    adc_raw.emit(50)
    await asyncio.sleep(2)

loop = asyncio.get_event_loop()
loop.run_until_complete(main())



""" Implementing Publisher """
from typing import (TYPE_CHECKING, TypeVar, Type, Tuple, Callable, Optional,
                    overload)

from broqer import NONE, Disposable
import broqer

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from typing import List
    from broqer import Subscriber


class SubscriptionError(ValueError):
    """ Special exception raised when subscription is failing (subscriber
    already subscribed) or on unsubscribe when subscriber is not subscribed
    """


TInherit = TypeVar('TInherit')  # Type to inherited behavior from
TValue = TypeVar('TValue')  # Type of publisher state and emitted value
SubscriptionCBT = Callable[[bool], None]


class Publisher:
    """ In broqer a subscriber can subscribe to a publisher. After subscription
    the subscriber is notified about emitted values from the publisher (
    starting with the current state). In other frameworks
    *publisher*/*subscriber* are referenced as *observable*/*observer*.

    broqer.NONE is used as default initialisation. .get() will always
    return the internal state (even when it's broqer.NONE). .subscribe() will
    emit the actual state to the new subscriber only if it is something else
    than broqer.NONE .

    To receive information use following methods to interact with Publisher:

    - ``.subscribe(subscriber)`` to subscribe for events on this publisher
    - ``.unsubscribe(subscriber)`` to unsubscribe
    - ``.get()`` to get the current state

    When implementing a Publisher use the following methods:

    - ``.notify(value)`` calls .emit(value) on all subscribers

    :param init: the initial state.

    :ivar _state: state of the publisher
    :ivar _inherited_type: type class for method lookup
    :ivar _subscriptions: holding a list of subscribers
    :ivar _on_subscription_cb: callback with boolean as argument, telling
                                if at least one subscription exists
    :ivar _dependencies: list with publishers this publisher is (directly or
                         indirectly) dependent on.
    """
    @overload  # noqa: F811
    def __init__(self, *, type_: Type[TValue] = None):
        pass

    @overload  # noqa: F811
    def __init__(self, init: TValue, type_: Type[TValue] = None):  # noqa: F811
        pass

    def __init__(self, init=NONE, type_=None):  # noqa: F811
        self._state = init

        if type_:
            self._inherited_type = type_  # type: Optional[Type]
        elif init is not NONE:
            self._inherited_type = type(init)
        else:
            self._inherited_type = None

        self._subscriptions = list()  # type: List[Subscriber]
        self._on_subscription_cb = None  # type: Optional[SubscriptionCBT]
        self._dependencies = ()  # type: Tuple[Publisher, ...]

    def subscribe(self, subscriber: 'Subscriber',
                  prepend: bool = False) -> 'SubscriptionDisposable':
        """ Subscribing the given subscriber.

        :param subscriber: subscriber to add
        :param prepend: For internal use - usually the subscribers will be
            added at the end of a list. When prepend is True, it will be added
            in front of the list. This will habe an effect in the order the
            subscribers are called.
        :raises SubscriptionError: if subscriber already subscribed
        """

        # `subscriber in self._subscriptions` is not working because
        # tuple.__contains__ is using __eq__ which is overwritten and returns
        # a new publisher - not helpful here
        if any(subscriber is s for s in self._subscriptions):
            raise SubscriptionError('Subscriber already registered')

        if not self._subscriptions and self._on_subscription_cb:
            self._on_subscription_cb(True)

        if prepend:
            self._subscriptions.insert(0, subscriber)
        else:
            self._subscriptions.append(subscriber)

        disposable_obj = SubscriptionDisposable(self, subscriber)

        if self._state is not NONE:
            subscriber.emit(self._state, who=self)

        return disposable_obj

    def unsubscribe(self, subscriber: 'Subscriber') -> None:
        """ Unsubscribe the given subscriber

        :param subscriber: subscriber to unsubscribe
        :raises SubscriptionError: if subscriber is not subscribed (anymore)
        """
        # here is a special implementation which is replacing the more
        # obvious one: self._subscriptions.remove(subscriber) - this will not
        # work because list.remove(x) is doing comparison for equality.
        # Applied to publishers this will return another publisher instead of
        # a boolean result
        for i, _s in enumerate(self._subscriptions):
            if _s is subscriber:
                self._subscriptions.pop(i)

                if not self._subscriptions and self._on_subscription_cb:
                    self._on_subscription_cb(False)

                return

        raise SubscriptionError('Subscriber is not registered')

    def get(self) -> TValue:
        """ Return the state of the publisher. """
        return self._state

    def notify(self, value: TValue) -> None:
        """ Calling .emit(value) on all subscribers and store state.

        :param value: value to be emitted to subscribers
        """
        self._state = value
        for subscriber in tuple(self._subscriptions):
            subscriber.emit(value, who=self)

    def reset_state(self) -> None:
        """ Resets the state. Calling this method will not trigger a
        notification, but will call .reset_state for all subscribers

        """
        self._state = NONE
        for subscriber in tuple(self._subscriptions):
            subscriber.reset_state()

    @property
    def subscriptions(self) -> Tuple['Subscriber', ...]:
        """ Property returning a tuple with all current subscribers """
        return tuple(self._subscriptions)

    def register_on_subscription_callback(self,
                                          callback: SubscriptionCBT) -> None:
        """ This callback will be called, when the subscriptions are changing.
        When a subscription is done and no subscription was present the
        callback is called with True as argument. When after unsubscribe no
        subscription is left, it will be called with False.

        :param callback: callback(subscription: bool) to be called.
                         when `callback` is None the callback will be reset
        :raises ValueError: when a callback is already registrered
        """
        if callback is None:
            self._on_subscription_cb = None
            return

        if self._on_subscription_cb is not None:
            raise ValueError('A callback is already registered')

        self._on_subscription_cb = callback
        if self._subscriptions:
            callback(True)

    def __await__(self):
        """ Makes publisher awaitable. When publisher has a state it will
        immediatly return its state as result. Otherwise it will wait forever
        until it will change its state.
        """
        future = self.as_future(timeout=None, omit_subscription=False)
        return future.__await__()

    def as_future(self, timeout: float, omit_subscription: bool = True,
                  loop=None):
        """ Returns a asyncio.Future which will be done on first change of this
        publisher.

        :param timeout: timeout in seconds. Use None for infinite waiting
        :param omit_subscription: if True the first emit (which can be on the
            subscription) will be ignored.
        :param loop: asyncio loop to be used
        :returns: a future returning the emitted value
        """
        return broqer.OnEmitFuture(self, timeout, omit_subscription, loop)

    def __bool__(self):
        """ A new Publisher is the result of a comparision between a publisher
        and something else (may also be a second publisher). This result should
        never be used in a boolean sense (e.g. in `if p1 == p2:`). To prevent
        this __bool__ is overwritten to raise a ValueError.
        """
        raise ValueError('Evaluation of comparison of publishers is not '
                         'supported')

    def __iter__(self):
        """ To prevent iterating over a publisher this method is implemented
        to throw an exception. Otherwise it will fallback to __getitem__.
        """
        raise ValueError('Iteration over a publisher is not possible')

    def inherit_type(self, type_cls: Optional[Type]) -> None:
        """ Enables the usage of method and attribute overloading for this
        publisher.
        """
        self._inherited_type = type_cls

    @property
    def inherited_type(self) -> Optional[Type]:
        """ Property inherited_type returns used type class (or None) """
        return self._inherited_type

    @property
    def dependencies(self) -> Tuple['Publisher', ...]:
        """ Returning a list of publishers this publisher is dependent on. """
        return self._dependencies

    def add_dependencies(self, *publishers: 'Publisher') -> None:
        """ Add publishers which are directly or indirectly controlling the
        behavior of this publisher

        :param *publishers: variable argument list with publishers
        """
        self._dependencies = self._dependencies + publishers

    def __dir__(self):
        """ Extending __dir__ with inherited type """
        attrs = set(super().__dir__())
        if self._inherited_type:
            attrs.update(set(dir(self._inherited_type)))
        return tuple(attrs)


class SubscriptionDisposable(Disposable):
    """ This disposable is returned on Publisher.subscribe(subscriber).
        :param publisher: publisher the subscription is made to
        :param subscriber: subscriber used for subscription
    """
    def __init__(self, publisher: 'Publisher', subscriber: 'Subscriber') \
            -> None:
        self._publisher = publisher
        self._subscriber = subscriber

    def dispose(self) -> None:
        self._publisher.unsubscribe(self._subscriber)

    @property
    def publisher(self) -> 'Publisher':
        """ Subscripted publisher """
        return self._publisher

    @property
    def subscriber(self) -> 'Subscriber':
        """ Subscriber used in this subscription """
        return self._subscriber



"""
Apply ``coro(*args, value, **kwargs)`` to each emitted value allowing async
processing.

Usage:

>>> import asyncio
>>> from broqer import Value, op
>>> s = Value()

>>> async def delay_add(a):
...     print('Starting with argument', a)
...     await asyncio.sleep(0.015)
...     result = a + 1
...     print('Finished with argument', a)
...     return result

MODE: CONCURRENT (is default)

>>> _d = s.subscribe(SinkAsync(delay_add))
>>> s.emit(0)
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument 0
Starting with argument 1
Finished with argument 0
Finished with argument 1
>>> _d.dispose()
"""

from functools import wraps
from typing import Any

# pylint: disable=cyclic-import
from broqer import Subscriber, Publisher, default_error_handler
from broqer.op.map_async import AppliedMapAsync, AsyncMode, build_coro


class SinkAsync(Subscriber):  # pylint: disable=too-few-public-methods
    """ Apply ``coro`` to each emitted value allowing async processing

    :param coro: coroutine to be applied on emit
    :param \\*args: variable arguments to be used for calling coro
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked as (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for calling coro
    """
    def __init__(self, coro, *args, mode=AsyncMode.CONCURRENT,
                 error_callback=default_error_handler,
                 unpack: bool = False, **kwargs) -> None:

        self._dummy_publisher = Publisher()
        _coro = build_coro(coro, unpack, *args, **kwargs)
        self._map_async = AppliedMapAsync(self._dummy_publisher,
                                          _coro,
                                          mode=mode,
                                          error_callback=error_callback)

    def emit(self, value: Any, who: Publisher):
        self._map_async.emit(value, who=self._dummy_publisher)


def build_sink_async(coro=None, *, mode: AsyncMode = AsyncMode.CONCURRENT,
                     unpack: bool = False):
    """ Decorator to wrap a coroutine to return a SinkAsync subscriber.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_sink_async(coro):
        return SinkAsync(coro, mode=mode, unpack=unpack)

    if coro:
        return _build_sink_async(coro)

    return _build_sink_async


def build_sink_async_factory(coro=None, *,
                             mode: AsyncMode = AsyncMode.CONCURRENT,
                             error_callback=default_error_handler,
                             unpack: bool = False):
    """ Decorator to wrap a coroutine to return a factory for SinkAsync
        subscribers.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_sink_async(coro):
        @wraps(coro)
        def _wrapper(*args, **kwargs) -> SinkAsync:
            if ('unpack' in kwargs) or ('mode' in kwargs) or \
                    ('error_callback' in kwargs):
                raise TypeError('"unpack", "mode" and "error_callback" has to '
                                'be defined by decorator')

            return SinkAsync(coro, *args, mode=mode,
                             error_callback=error_callback, unpack=unpack,
                             **kwargs)
        return _wrapper

    if coro:
        return _build_sink_async(coro)

    return _build_sink_async


def sink_async_property(coro=None, *,
                        mode: AsyncMode = AsyncMode.CONCURRENT,
                        error_callback=default_error_handler,
                        unpack: bool = False):
    """ Decorator to build a property returning a SinkAsync subscriber.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    def build_sink_async_property(coro):
        @property
        def _build_sink_async(self):
            return SinkAsync(coro, self, mode=mode,
                             error_callback=error_callback, unpack=unpack)
        return _build_sink_async

    if coro:
        return build_sink_async_property(coro)

    return build_sink_async_property



from unittest import mock
from itertools import product

import pytest

from broqer import Publisher, NONE, op, Sink
from tests.helper_multi import check_get_method, check_subscription, check_dependencies


test_vector = [
    # o, args, kwargs, input_vector, output_vector
    (op.CombineLatest, (), {},
        ((1, 2), (NONE, 3), (2, NONE), (2, 3)),
        ((1, 2), (1, 3), (2, 3), (2, 3), (2, 3))),
    (op.CombineLatest, (), {},
        ((1,), (2,), (2,), (3,)),
        (((1,), (2,), (2,), (3,)))),
    (op.CombineLatest, (), {},
        ((1, NONE, NONE, NONE), (NONE, 2, NONE, NONE), (NONE, 3, NONE, NONE), (NONE, NONE, 4, 5)),
        ((NONE, NONE, NONE, (1, 3, 4, 5)))),
    (op.CombineLatest, (), {'map_': lambda a, b: a+b},
        ((1, 1), (NONE, 2), (1, NONE), (NONE, -5)),
        (2, 3, 3, -4)),
    (op.CombineLatest, (), {'map_': lambda a, b: a > b},
        ((1, 1), (NONE, 2), (1, NONE), (NONE, -5)),
        (False, False, False, True)),
    (op.CombineLatest, (), {'map_': lambda a, b: NONE if a > b else a - b},
        ((0, 0), (NONE, 1), (1, NONE), (NONE, 0), (0, NONE)),
        (0, -1, 0, NONE, 0)),
    (op.build_combine_latest(lambda a, b: a + b), (), {},
        ((1, 1), (NONE, 2), (1, NONE), (NONE, -5)),
        (2, 3, 3, -4)),
    (op.build_combine_latest(map_=lambda a, b: a + b), (), {},
        ((1, 1), (NONE, 2), (1, NONE), (NONE, -5)),
        (2, 3, 3, -4)),
    (lambda p1, p2, f: op.build_combine_latest()(f)(p1, p2), (lambda a, b: a + b,), {},
        ((1, 1), (NONE, 2), (1, NONE), (NONE, -5)),
        (2, 3, 3, -4)),
]
@pytest.mark.parametrize('method', [check_get_method, check_subscription, check_dependencies])
@pytest.mark.parametrize('o,args,kwargs,input_vector,output_vector', test_vector)
def test_operator(method, o, args, kwargs, input_vector, output_vector):
    operator = o(*(Publisher(v) for v in input_vector[0]), *args, **kwargs)

    method(operator, input_vector, output_vector)


@pytest.mark.parametrize('factory', [
    lambda p1, p2, p3, emit_on: op.CombineLatest(p1, p2, p3, emit_on=emit_on),
    lambda p1, p2, p3, emit_on: op.build_combine_latest(lambda *n: n, emit_on=emit_on)(p1, p2, p3),
    ])
@pytest.mark.parametrize('flags', list(product([False, True], repeat=3)))
def test_emit_on(factory, flags):
    m = mock.Mock()

    publishers = [Publisher(None) for i in range(3)]

    emit_on = [p for p, f in zip(publishers, flags) if f]
    if len(emit_on) == 0:
        emit_on = None
    elif len(emit_on) == 1:
        emit_on = emit_on[0]

    operator = factory(*publishers, emit_on=emit_on)
    operator.subscribe(Sink(m))

    result = [None] * 3

    if flags.count(True) == 0:
        flags = [True] * 3

    for i, f in enumerate(flags):
        m.reset_mock()

        result[i] = i

        publishers[i].notify(i)

        if f:
            m.assert_called_once_with(tuple(result))
        else:
            m.assert_not_called()

@pytest.mark.parametrize('subscribe', [True, False])
def test_unsubscibe(subscribe):
    m = mock.Mock()

    p1, p2 = Publisher(NONE), Publisher(NONE)

    operator = op.CombineLatest(p1, p2)

    if subscribe:
        operator.subscribe(Sink())

    assert operator.get() == NONE

    p1.notify(1)
    assert operator.get() == NONE

    p2.notify(2)
    assert operator.get() == (1, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" The setup script."""

from setuptools import find_packages, setup


with open('README.rst', 'rb') as readme_file:
    readme = readme_file.read().decode('utf-8')


setup(
    author='Günther Jena',
    author_email='guenther@jena.at',
    use_scm_version={"write_to": "broqer/_version.py"},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description='Carefully crafted library to operate with continuous ' +
                'streams of data in a reactive style with publish/subscribe ' +
                'and broker functionality.',
    license='MIT license',
    long_description=readme,
    include_package_data=True,
    keywords='broker publisher subscriber reactive frp observable',
    name='broqer',
    packages=find_packages(include=['broqer*']),
    url='https://github.com/semiversus/python-broqer',
    zip_safe=False,
)



import pytest

from broqer import op, NONE, Value
from tests.helper_single import check_get_method, check_subscription, \
                                check_dependencies


test_vector = [
    # o, args, kwargs, input_vector, output_vector
    (op.Filter, (lambda v: v == 0,), {},
        (1, 2, 0, 0.0, None),
        (NONE, NONE, 0, 0.0, NONE)),
     (op.Filter, (lambda v: v == 0,), {},
        (0, ),
        (0, )),
    (op.Filter, (lambda a, b: a == b, 2), {},
        (0, 2, 2.0, 1),
        (NONE, 2, 2.0, NONE)),
    (op.Filter, (lambda a, b, c: a + b == c, 2), {'unpack': True},
        ((0, 2), (0, 3), (-1.0, 1.0)),
        ((0, 2), NONE, (-1.0, 1.0))),
    (op.Filter, (lambda a, b, c: a + b == c, 2), {'unpack': True},
        ((0, 3),),
        (NONE,)),
    (op.Filter, (lambda a, b, c: a + b == c,), {'unpack': True, 'c': 2},
        ((0, 2), (0, 3), (-1.0, 3.0)),
        ((0, 2), NONE, (-1.0, 3.0))),
    (op.build_filter_factory(lambda v: v == 0), (), {},
        (1, 2, 0, 0.0, None),
        (NONE, NONE, 0, 0.0, NONE)),
    (op.build_filter_factory(unpack=False)(lambda v: v == 0), (), {},
        (1, 2, 0, 0.0, None),
        (NONE, NONE, 0, 0.0, NONE)),
    (op.build_filter_factory(lambda a, b, c: a + b == c, unpack=True), (2,), {},
        ((0, 2), (0, 3), (-1.0, 1.0)),
        ((0, 2), NONE, (-1.0, 1.0))),
    (lambda: op.build_filter(lambda v: v == 0), (), {},
        (1, 2, 0, 0.0, None),
        (NONE, NONE, 0, 0.0, NONE)),
    (lambda: op.build_filter(unpack=False)(lambda v: v == 0), (), {},
        (1, 2, 0, 0.0, None),
        (NONE, NONE, 0, 0.0, NONE)),
]


@pytest.mark.parametrize('method', [check_get_method, check_subscription,
                                    check_dependencies])
@pytest.mark.parametrize('o,args,kwargs,input_vector,output_vector',
                         test_vector)
def test_operator(method, o, args, kwargs, input_vector, output_vector):
    operator = o(*args, **kwargs)

    method(operator, input_vector, output_vector)


test_vector = [
    # o, input_vector, output_vector
    (op.EvalTrue,
        (1, 2, 0, 0.0, None, False, [1]),
        (1, 2, NONE, NONE, NONE, NONE, [1])),
    (op.EvalTrue,
        (0, ),
        (NONE, )),
    (op.EvalFalse,
        (1, 2, 0, 0.0, None, False, [1]),
        (NONE, NONE, 0, 0.0, None, False, NONE)),
    (op.EvalFalse,
        (0, ),
        (0, )),
]


@pytest.mark.parametrize('method', [check_get_method, check_subscription,
                                    check_dependencies])
@pytest.mark.parametrize('o,input_vector,output_vector',
                         test_vector)
def test_true_false(method, o, input_vector, output_vector):
    method(o, input_vector, output_vector)


def test_filter_factory_keyword():
    m = op.build_filter_factory(lambda v: v)
    v = Value()

    with pytest.raises(TypeError, message='"unpack" has to be defined by decorator'):
        o = v | m(unpack=True)


"""
Build a future able to await for

Usage:

>>> import asyncio
>>> from broqer import Value, op, OnEmitFuture
>>> s = Value()

>>> _ = asyncio.get_event_loop().call_later(0.05, s.emit, 1)

>>> asyncio.get_event_loop().run_until_complete(OnEmitFuture(s) )
1

#>>> _ = asyncio.get_event_loop().call_later(0.05, s.emit, (1, 2))
#>>> asyncio.get_event_loop().run_until_complete(s)
(1, 2)
"""
import asyncio
from typing import Any, Optional, TYPE_CHECKING

import broqer

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from broqer import Publisher


class OnEmitFuture(broqer.Subscriber, asyncio.Future):
    """ Build a future able to await for.
    :param publisher: source publisher
    :param timeout: timeout in seconds
    :param loop: asyncio loop to be used
    """
    def __init__(self, publisher: 'Publisher', timeout=None,
                 omit_first_emit=False, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()

        asyncio.Future.__init__(self, loop=loop)
        self.add_done_callback(self._cleanup)

        self._publisher = publisher

        self._omit_first_emit = omit_first_emit

        if timeout is not None:
            self._timeout_handle = loop.call_later(
                timeout, self.set_exception, asyncio.TimeoutError)
        else:
            self._timeout_handle = None

        publisher.subscribe(self)

    def _cleanup(self, _future=None):
        self._publisher.unsubscribe(self)

        if self._timeout_handle is not None:
            self._timeout_handle.cancel()
            self._timeout_handle = None

    def emit(self, value: Any, who: Optional['Publisher'] = None) -> None:
        if who is not self._publisher:
            raise ValueError('Emit from non assigned publisher')

        if self._omit_first_emit:
            self._omit_first_emit = False
            return

        if not self.done():
            self.remove_done_callback(self._cleanup)
            self._cleanup()
            self.set_result(value)



""" Implementing DefaultErrorHandler. Object default_error_handler is used
as global object to register a callbacks for exceptions in asynchronous
operators, """
import traceback


def _default_error_callback(exc_type, exc_value, exc_traceback):
    """ Default error callback is printing traceback of the exception
    """
    traceback.print_exception(exc_type, exc_value, exc_traceback)


class DefaultErrorHandler:
    """ DefaultErrorHandler object is a callable which is calling a registered
    callback and is used for handling exceptions when asynchronous operators
    receiving an exception during .emit(). The callback can be registered via
    the .set(callback) method. The default callback is _default_error_callback
    which is dumping the traceback of the exception.
    """
    def __init__(self):
        self._error_callback = _default_error_callback

    def __call__(self, exc_type, exc_value, exc_traceback):
        """ When calling the call will be forwarded to the registered
        callback """
        self._error_callback(exc_type, exc_value, exc_traceback)

    def set(self, error_callback):
        """ Register a new callback

        :param error_callback: the callback to be registered
        """
        self._error_callback = error_callback

    def reset(self):
        """ Reset to the default callback (dumping traceback)
        """
        self._error_callback = _default_error_callback


default_error_handler = DefaultErrorHandler()  # pylint: disable=invalid-name
